// alloc 1024 memory blocks and test for errors
//   free good memory blocks
//   do not free bad memory blocks and run (+alloc bad blocks) forever
// if some bits are bad, 0.1% of memory will be occupied and rest can be used
//
// gcc -O2 -o mem_badram mem_badram.c
// ./mem_badram 31000  # argument1 is physical memory in MB
//
// 2011-07-06 cosmetic changes for output (more info)
// 2013 better use memspeed --test=8 or 16

#include <stdlib.h>
#include <stdio.h>

// analyse memory given by pointer and length in sizeoflong
int chkmem(long int *a, long int num) {
  long int jj=0, i, j, rnd;
  for (j=1;j<8; j++){
    rnd=rand();     /* 2013-03 ToDo: ohne srand immer gleich (srand(1))!
                       besser 1k zufall (rnd[i&1023] kopieren? srand per argument? */
    for (i=0; i<num; i++) a[i]=rnd;
    for (i=0; i<num; i++) if (a[i]!=rnd) {
      printf("ERROR bad data at a[%08x]=%08x != %08x diff=%08x (j=%d)\n",
        i,(unsigned long)a[i],(unsigned long)rnd,(unsigned long)(a[i]^rnd),j); 
        jj=i; break;
    }
    if (jj) break;
    for (i=0; i<num; i++) a[i]=~rnd;
    for (i=0; i<num; i++) if (a[i]!=~rnd) {
      printf("ERROR bad data at a[%08x]=%08x != %08x diff=%08x (j=%d)\n",
        i,(unsigned long)a[i],(unsigned long)~rnd,(unsigned long)(a[i]^~rnd),j);
        jj=i; break;
    }
    if (jj) break;
  }
  return jj;
}

int main(int argn, char *argv[]){
  long int rnd, i, j, *pa[1024], bad=0, num, jj, nb=0, mb=1024;
  int sl=sizeof(long int);
  if (argn>1) mb=atoi(argv[1]);
  num = mb*(1024l/sizeof(long int));
  printf(" malloc %d MiB (Mi=2^20) long= %d B = %d bit\n", mb, sl, sl*8);
  // split into 1024 blocks of size mb*1024
  printf(" malloc 1024 blocks to %d kiB = %d * long (ki=2^10)\n", mb, num);
  for (i=0;i<1024;i++) {
    pa[i] = (long int *) malloc(mb*1024l);
    if ((i&127)==0)printf(" a=%012lx blk %d (malloc+touch/512)\n", pa[i], i);
    if (!pa[i]) break;
    for (j=0; j<num; j+=512) pa[i][j]=0; // touch every 512 B!
  } nb=i;
  if (nb!=1024)
    printf(" WARNING: only %d blocks = %d kiB allocated\n", nb+1, (nb+1)*mb);
  if (!nb) exit(1);
  for (i=0;i<nb;i++) {
    if ((i&127)==0)printf(" a=%012lx blk %d (8wr~wr)\n", pa[i], i);
    if (pa[i]) jj = chkmem(pa[i],num); // 0=OK
    if (jj) { bad++; pa[i]=0; } // mark as bad
  }
  for (i=0;i<nb;i++) if (pa[i]) free(pa[i]); // free good blocks
  if (bad) for (;;) system("sleep 600"); // sleep forever
  return 0;
}
